#include <math.h>
#include <limits.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/resource.h>

#define PRIVATE static
#define SVMZERO ((double) 0)
#define SVMONE ((double) 1)
#define SVMMIN(a,b) (((a) < (b)) ? (a) : (b))
#define SVMMAX(a,b) (((a) > (b)) ? (a) : (b))
#define SVMINF DBL_MAX
#define SVMINT_INF INT_MAX
#define SVMTRUE 1
#define SVMFALSE 0

/* include the user-visible definitions */
#include "SVM_user.h"

typedef struct SVMcom_struct /* SVM variables */
{
    int             n ; /* problem dimension */
    short      kernel ; /* 0 = linear, 1 = poly, 2 = rbf, 3 = sig */
    double        tol ; /* tolerance = KKT violation */
    double          C ; /* SVM constraint upper bound */
    double          b ; /* SVM right side for linear constraint 1'x = b */
    int          bsub ; /* bound on size of subspace */
    short      *ikeep ; /* array of size n initialized to 0 */
    float         *Ax ; /* numeric values */
    double    *normAx ; /* norm squared of a data column, columns are float */
    short     *normAi ; /* norm squared of a data column, columns are 1s  */
    short         *Ai ; /* features */
    int           *Ap ; /* pointers */
    int      nfeature ; /* number of features */
    SVMparm     *Parm ; /* SVM parameter structure */
    float       *Sold ; /* sparse to store old copy of subspace matrix S */
    float     *Astore ; /* storage for evaluated columns of A */
    double *xWorkUpdate; /* double space needed in gradient update */
    int    *iWorkUpdate; /* int    space needed in gradient update */
    float  *fWorkUpdate; /* float  space needed in gradient update */
    double *xWorkSubspace;/* double space needed in subspace computation */
    int    *iWorkSubspace;/* int    space needed in subspace computation */
} SVMcom ;

typedef struct SVMmatrices /* SVM matrices */
{
    float     *Sdense ; /* dense version of subspace matrix S */
    float         *Sx ; /* numerical values for S'S version of S */
    int           *Sp ; /* column pointers */
    short         *Si ; /* row indices */
    double      *temp ; /* temporary storage of size nfeature */
    int      nfeature ; /* number of features */
    int         FormS ; /* TRUE means that the dense S is constructed */
    short     Ax_is_1 ; /* 1 means all element of Ax are 1, Ax ignored */
} SVMmatrices ;

/* prototypes */

PRIVATE void SVMsparseProduct
(
    double *p,
    double *d,
    int     n
) ;

PRIVATE void SVMSmult
(
    BLGobjective *data
) ;

PRIVATE int SVMbuildSubspace /* return:
                   0 (error tolerance satisfied)
                   1 (refine accuracy for prior subspace)
                   2 (subspace has been evaluated) */
(
    int          *i, /* indices of subspace (returned) */
    int       *nsub, /* size of subspace (returned) */
    double     *Err, /* error estimate (returned) */
    double       *B, /* 1'x outside subsubspace (returned) */
    double      *nu, /* multiplier estimate (input and returned) */
    int     *istore, /* stored column in current subspace */
    int     *nstore, /* number of stored columns in current subspace */
    int      *ifree, /* free indices from previous subspace */
    int       nfree, /* number of free indices from previous subspace */
    double   SubErr, /* error for previous subspace */
    double       *x, /* current estimate of solution to problem */
    double       *g, /* gradient at current point */
    double lambda_k, /* current BB parameter */
    SVMcom     *Com  /* SVM Com structure */
) ;

PRIVATE void SVMupdateGrad
(
    double       *g, /* gradient to be updated */
    double       *d, /* change in the subspace: gnew = g + A*d */
    double       *y, /* solution to subspace problem */
    int          *I, /* indices or columns in subspace */
    int        nsub, /* subspace size */
    int     *istore, /* indices of stored columns for subspace */
    int      nstore, /* number stored columns in subspace */
    int      *ifree, /* indices for free variables in subspace */
    int      *nfree, /* number free variables in subspace */
    float        *S, /* subspace matrix */
    float       *Sx, /* numerical data */
    short       *Si, /* features associated with data */
    int         *Sp, /* pointer array */
    float     *Colx, /* storage for float column of data */
    short     *Coli, /* storage for 1s column of data */
    double  *NormAx, /* norm of each column, columns are float */
    short   *NormAi, /* norm of each column, columns are 1s */
    SVMcom     *Com  /* SVM Com structure */
) ;

PRIVATE void SVMbuildS
(
    float       *S, /* svm matrix corresponding to subspace */
    float      *Sx, /* numerical data */
    short      *Si, /* features associated with data */
    int        *Sp, /* pointer array */
    float    *Colx, /* storage for float column of data */
    short    *Coli, /* storage for 1s column of data */
    double *NormAx, /* norm of each column, columns are float */
    short  *NormAi, /* norm of each column, columns are 1s */
    int         *I, /* subspace indices */
    int       nsub, /* size of I */
    int      *Iold, /* indices for prior subspace */
    int       nold, /* size of Iold */
    SVMcom    *Com
) ;

double SVMdouble_dense_dot
(
    double *x,  /* nonzero numerical values, dense format */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* location in x of nonzero indices in y */
    int     n   /* number of nonzeros in y */
) ;

/* dot product x'y where x is dense and y is sparse */
double SVMdense_sparse_dot
(
    double *x,  /* nonzero numerical values, dense format */
    float  *y,  /* nonzero numerical values */
    int   *yi,  /* location in y of nonzero indices in x */
    int     n   /* number of nonzeros in y */
) ;

double SVMfloat_dense_dot
(
    float  *x,  /* nonzero numerical values, dense format */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int     n   /* number of nonzeros in y */
) ;

double SVMbinary_double_dot
(
    double *x,  /* nonzero numerical values, dense format */
    short *yi,  /* nonzero indices in y */
    int     n   /* number of nonzeros in y */
) ;

short SVMbinary_short_dot
(
    short  *x,  /* nonzero numerical values, dense format */
    short *yi,  /* nonzero indices in y */
    int     n   /* number of nonzeros in y */
) ;

double SVMsparse_dot
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

short SVMbinary_sparse_dot
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

double SVMfloatxx
(
    float  *x,  /* nonzero numerical values */
    int     n   /* number of nonzeros in x */
) ;

PRIVATE double SVMfloat_double_dot
(
    float  *x, /* first vector */
    double *y, /* second vector */
    int     n  /* length of vectors */
) ;

PRIVATE void SVMshort_gather
(
    short *y, /* compressed vector */
    short *x, /* current x */
    int   *i, /* indices associated with subspace */
    int    n  /* number of elements to extract from x */
) ;

PRIVATE void SVMdouble_gather
(
    double *y, /* compressed vector */
    double *x, /* current x */
    int    *i, /* indices associated with subspace */
    int     n  /* number of elements to extract from x */
) ;

PRIVATE void SVMfloat_gather
(
    float *y, /* compressed vector */
    float *x, /* current x */
    int   *i, /* indices associated with subspace */
    int    n  /* number of elements to extract from x */
) ;

PRIVATE void SVMdouble_scatter
(
    double *x, /* scatter y into x */
    double *y, /* compressed vector */
    int    *i, /* indices associated with subspace */
    int     n  /* number of elements to extract from x */
) ;

PRIVATE void SVMfloat_scatter
(
    float *x, /* scatter y into x */
    float *y, /* compressed vector */
    int   *i, /* indices associated with subspace */
    int    n  /* number of elements to extract from x */
) ;

PRIVATE void SVMdoubleCopy
(
    double *x, /* copy of y */
    double *y, /* given vector */
    int     n  /* length of vectors */
) ;

PRIVATE void SVMfloatCopy
(
    float *x, /* copy of y */
    float *y, /* given vector */
    int    n  /* length of vectors */
) ;

PRIVATE void SVMintCopy
(
    int *x, /* copy of y */
    int *y, /* given vector */
    int  n  /* length of vectors */
) ;

PRIVATE void SVMshortCopy
(
    short *x, /* copy of y */
    short *y, /* given vector */
    int    n  /* length of vectors */
) ;

PRIVATE void SVMindexCopy
(
    float    *x, /* target of copy is x */
    float    *y, /* copy from y */
    int   *iget, /* location in y to get element */
    int   *iput, /* location in x to put element */
    int       n  /* length of vectors */
) ;

PRIVATE void SVMindexCopyFlag
(
    float    *x, /* target of copy is x */
    float    *y, /* copy from y */
    int   *iget, /* location in y to get element */
    int   *iput, /* location in x to put element */
    int       n, /* length of vectors */
    short  flag  /* = 0 for collision free copy, +-1 when collisions possible */
) ;

PRIVATE void SVMdense_diff
(
    double    *x,
    double    *y,
    double    *z,
    int        n  /* dimension */
) ;

PRIVATE void SVMsparse_add
(
    double    *x,
    short    *yi,
    double     s, /* scalar */
    int        n  /* dimension */
) ;

PRIVATE void SVMsparse_daxpy
(
    double    *x,
    float     *y,
    short    *yi,
    double     s, /* scalar */
    int        n  /* dimension */
) ;

PRIVATE void SVMdense_daxpy
(
    double    *x,
    float     *y,
    double     s, /* scalar */
    int        n  /* dimension */
) ;

PRIVATE void SVMint_init
(
    short *x,  /* array to be initialized */
    short  s,  /* scalar */
    int    n   /* length of x */
) ;

PRIVATE void SVMscalar_init
(
    double  *x,  /* numerical values */
    double   s,  /* scalar */
    int      n   /* length of vectors */
) ;

PRIVATE double SVMadd
(
    double *x, /* vector */
    int     n  /* length of vector */
) ;

PRIVATE void SVMstep
(
    double *xnew, /* updated x vector */
    double    *x, /* current x */
    double    *d, /* search direction */
    double    st, /* stepsize */
    int        n  /* dimension */
) ;

PRIVATE double SVMmax
(
    double *x,
    int     n
) ;

PRIVATE void SVMpartialMinSort
(
    int     *i, /* indices of m smallest numbers */
    double  *x, /* numbers to sort */
    int      m, /* length of i */
    int      n  /* length of x */
) ;

PRIVATE void SVMpartialMaxSort
(
    int    *i, /* indices of m largest numbers */
    double *x, /* numbers to sort */
    int     m, /* length of i */
    int     n  /* length of x */
) ;

PRIVATE void SVMminHeapSort
(
    int     *i, /* indices with decreasing order based on x */
    int  *heap, /* min heap */
    double  *x, /* numbers to sort */
    int       n  /* size of heap */
) ;

PRIVATE void SVMminheap_build
(
    int  *heap, /* on input, an unsorted set of indices */
    double  *x, /* numbers to sort */
    int      n  /* number of elements to build into the heap */
) ;

PRIVATE int SVMminheap_delete  /* return new size of heap */
(
    int  *heap, /* containing indices into x, 1..n on input */
    double  *x, /* not modified */
    int      n  /* number of items in heap */
) ;

PRIVATE int SVMminheap_add
(
    int   leaf, /* the new leaf */
    int  *heap, /* size n, containing indices into x */
    double  *x, /* not modified */
    int      n  /* number of elements in heap not counting new one */
) ;

PRIVATE void SVMminheapify
(
    int *heap, /* size n, containing indices into x */
    double *x, /* not modified */
    int     p, /* start at node p in the heap */
    int     n  /* heap [1 ... n] is in use */
) ;

PRIVATE void SVMmaxheap_build
(
    int  *heap, /* on input, an unsorted set of elements */
    double  *x,
    int      n  /* number of elements to build into the heap */
) ;

PRIVATE int SVMmaxheap_delete  /* return new size of heap */
(
    int  *heap, /* containing indices into x, 1..n on input */
    double  *x, /* not modified */
    int      n  /* number of items in heap */
) ;

PRIVATE int SVMmaxheap_add
(
    int   *heap, /* size n, containing indices into x */
    double   *x, /* not modified */
    int    leaf, /* the new leaf */
    int       n  /* number of elements in heap not counting new one */
) ;

PRIVATE void SVMmaxheapify
(
    int *heap, /* size n, containing indices into x */
    double *x, /* not modified */
    int     p, /* start at node p in the heap */
    int     n  /* heap [1 ... n] is in use */
) ;

PRIVATE double SVMnapsack /* return the final lambda */
(
    double         *x, /* holds y on input, and the solution x on output */
    double     lambda, /* starting guess for lambda */
    double          b, /* right side */
    double         lo, /* lower bounds */
    double         hi, /* upper bounds */
    int             n, /* problem dimension */
    double         *w, /* work array of length n */
    int          *iw1, /* work array of length n */
    int          *iw2  /* work array of length n */
) ;

PRIVATE double SVMnapup /* return lambda */
(
    double        *y, /* input is target and output is solution */
    double         b, /* right side */
    double        lo, /* lower bounds */
    double        hi, /* upper bounds */
    int            n, /* problem dimension */
    double    lambda, /* initial guess for multiplier */
    double       atx, /* a'x */
    double     a2sum, /* sum_j a_j^2 */
    double *breakpts, /* break points */
    int  *bound_heap, /* work array */
    int   *free_heap  /* work array */
) ;

PRIVATE double SVMnapdown /* return lambda */
(
    double        *y, /* input is target and output is solution */
    double         b, /* right side */
    double        lo, /* lower bounds */
    double        hi, /* upper bounds */
    int            n, /* problem dimension */
    double    lambda, /* initial guess for multiplier */
    double       atx, /* a'x */
    double     a2sum, /* sum_j a_j^2 */
    double *breakpts, /* break points */
    int  *bound_heap, /* work array */
    int   *free_heap  /* work array */
) ;

PRIVATE double SVMlp /* return correct multiplier for linear constraint */
(
    double  *x,
    double  *c,
    double   C,
    double   b,
    int      n,
    double  mu, /* guess for multiplier */
    int     *i, /* work array of size n */
    int *index, /* work array of size n */
    double  *y  /* work array of size n */
) ;

PRIVATE void SVMminSort
(
    int *x,  /* numbers to sort (input), sorted numbers (output) */
    int *y,  /* work array of length n */
    int *z,  /* work array of length n+1 */
    int  n   /* number of elements to sort */
) ;

PRIVATE void SVMprint_parms
(
    SVMparm     *Parm /* SSMparm structure to be printed */
) ;

PRIVATE double SVMtimer ( void ) ;
